{ ##
  @FILE                     PJCBView.pas
  @COMMENTS                 Clipboard Viewer Component source file.
  @PROJECT_NAME             Clipboard Viewer Component
  @PROJECT_DESC             Component that registers itself as a clipboard
                            viewer to get notifications of when clipboard
                            contents change.
  @AUTHOR                   Peter Johnson, LLANARTH, Ceredigion, Wales, UK
  @EMAIL                    peter.johnson@openlink.org
  @WEBSITE                  http://www.delphidabbler.com/
  @COPYRIGHT                 Peter D Johnson, 1999-2004.
  @LEGAL_NOTICE             These components are distributed under the Mozilla
                            Public License - see below.
  @OTHER_NAMES              + Original file name was CBView.pas
                            + Changed to PJCBView.pas at v1.1
  @HISTORY(
    @REVISION(
      @VERSION              1.0
      @DATE                 01/08/1999
      @COMMENTS             Original version.
    )
    @REVISION(
      @VERSION              1.1
      @DATE                 27/07/2003
      @COMMENTS             + Changed component palette from "PJ Stuff" to
                              "DelphiDabbler".
                            + Moved ClipboardChanged method (that triggers
                              OnClipboardChanged event) from public to protected
                              scope and made dynamic rather than static. This
                              was done to make it easier to sub-class the
                              component.
    )
    @REVISION(
      @VERSION              1.2
      @DATE                 11/05/2004
      @COMMENTS             Altered to compile without deprecation warnings on
                            Delphi 6 and Delphi 7 by using conditional
                            compilation to using AllocateHWnd and DeallocateHWnd
                            from Classes rather than Forms unit.
    )
  )
}


{
 * ***** BEGIN LICENSE BLOCK *****
 * 
 * Version: MPL 1.1
 * 
 * The contents of this file are subject to the Mozilla Public License Version
 * 1.1 (the "License"); you may not use this file except in compliance with the
 * License. You may obtain a copy of the License at http://www.mozilla.org/MPL/
 * 
 * Software distributed under the License is distributed on an "AS IS" basis,
 * WITHOUT WARRANTY OF ANY KIND, either express or implied. See the License for
 * the specific language governing rights and limitations under the License.
 * 
 * The Original Code is Clipboard Viewer Component.
 * 
 * The Initial Developer of the Original Code is Peter Johnson
 * (http://www.delphidabbler.com/).
 * 
 * Portions created by the Initial Developer are Copyright (C) 1999-2004 Peter
 * Johnson. All Rights Reserved.
 * 
 * Contributor(s):
 * 
 * ***** END LICENSE BLOCK *****
}


unit PJCBView;


interface


{ Determine compiler }
{$IFDEF VER140}
  {$DEFINE DELPHI6ANDUP}
{$ENDIF}
{$IFDEF VER150}
  {$DEFINE DELPHI6ANDUP}
  {$DEFINE DELPHI7ANDUP}
{$ENDIF}



uses
  {Delphi}
  {$IFDEF WIN32}
    Windows,
  {$ELSE}
    WinProcs, WinTypes,
  {$ENDIF}
  Messages, Classes;


type

  {
  TPJCBViewer:
    Component that monitors Windows clipboad and triggers an event whenever
    clipboard contents change.
  }
  TPJCBViewer = class(TComponent)
  private {properties}
    fOnClipboardChanged: TNotifyEvent;
    fTriggerOnCreation: Boolean;
    fEnabled: Boolean;
  private
    fHWnd: HWND;
      {Handle of hidden window used as a clipboard viewer}
    fHWndNextViewer: HWND;
      {Handle of next clipboard viewer window in chain}
  protected
    procedure ClipboardChanged; dynamic;
      {Triggers OnClipboardChanged event if any handler is assigned and
      component is enabled}
    procedure Loaded; override;
      {Overrride of Loaded method to fire OnClipboardChanged event if required
      on component creation}
    procedure WndMethod(var Msg: TMessage); virtual;
      {Window procedure for hidden window}
  public
    constructor Create(AOwner: TComponent); override;
      {Class constructor: creates a hidden window, registers it as a clipboard
      viewer and sets default property values}
    destructor Destroy; override;
      {Class destructor: de-registers hidden window as clipboard viewer and
      disposes of it}
  published
    property OnClipboardChanged: TNotifyEvent
      read fOnClipboardChanged write fOnClipboardChanged;
      {Event triggered whenever clipboard contents change}
    property TriggerOnCreation: Boolean
      read fTriggerOnCreation write fTriggerOnCreation default True;
      {When true causes OnClipboardChanged event to be triggered as soon as
      component is created, otherwise OnClipboardChanged event only triggers
      when clipboard actually changes}
    property Enabled: Boolean read fEnabled write fEnabled default True;
      {When true component triggers events when clipboard changes, when false
      these events are not triggered}
  end;


procedure Register;
  {Component registration procedure}


implementation


uses
  {Delphi}
  Forms;


procedure TPJCBViewer.ClipboardChanged;
  {Triggers OnClipboardChanged event if any handler is assigned and component is
  enabled}
begin
  try
    {Fire event if assigned and enabled}
    if Assigned(fOnClipboardChanged) and fEnabled then
      fOnClipboardChanged(Self);
  except
    {Ensure any un-caught exception is handled by application and then continue}
    Application.HandleException(Self);
  end;
end;

constructor TPJCBViewer.Create(AOwner: TComponent);
  {Class constructor: creates a hidden window, registers it as a clipboard
  viewer and sets default property values}
begin
  inherited Create(AOwner);
  {Create hidden window to be used as clipboard viewer}
  {$IFDEF DELPHI6ANDUP}
    fHWnd := Classes.AllocateHWnd(WndMethod);
  {$ELSE}
    fHWnd := Forms.AllocateHWnd(WndMethod);
  {$ENDIF}
  {Register hidden window as a clipboard viewer and store handle of next window
  in chain}
  fHWndNextViewer := SetClipboardViewer(fHWnd);
  {Set default property values}
  fTriggerOnCreation := True;
  fEnabled := True;
end;

destructor TPJCBViewer.Destroy;
  {Class destructor: de-registers hidden window as clipboard viewer and disposes
  of it}
begin
  {Remove our hidden window from the chain of clipboard viewers}
  ChangeClipboardChain(fHWnd, fHWndNextViewer);
  {Get rid of the window}
  {$IFDEF DELPHI6ANDUP}
    Classes.DeallocateHWnd(fHWnd);
  {$ELSE}
    Forms.DeallocateHWnd(fHWnd);
  {$ENDIF}
  inherited Destroy;
end;

procedure TPJCBViewer.Loaded;
  {Overrride of Loaded method to fire OnClipboardChanged event if required on
  component creation}
begin
  inherited Loaded;
  {If we're notifying a clipboard change event on creation then do it}
  if fTriggerOnCreation then
    ClipboardChanged;
end;

procedure TPJCBViewer.WndMethod(var Msg: TMessage);
  {Window procedure for hidden window}
var
  Handled: Boolean; {flag true if we handle a message}
begin
	{Assume we've handled message}
  Handled := True;
  {Now process messages, handling the ones we want}
  with Msg do
  begin
  	case Msg of
      WM_DRAWCLIPBOARD:
      begin
        {Clipboard has changed: trigger event}
        ClipboardChanged;
        {Pass on message to next window in viewer chain (if any)}
        if fHWndNextViewer <> 0 then
          SendMessage(fHWndNextViewer, Msg, WParam, LParam);
      end;
      WM_CHANGECBCHAIN:
      begin
        {Windows is detaching a clipboard viewer}
        if HWND(WParam) = fHWndNextViewer then
          {Window being detached is next one: record new "next" window in chain}
          fHWndNextViewer := HWND(LParam)
        else
          {Window being detached is not next one: pass message along}
          if fHWndNextViewer <> 0 then
            SendMessage(fHWndNextViewer, Msg, WParam, LParam);
      end;
      else
      begin
        {We're not handling this message}
        Handled := False;
      end;
    end;
    {Check if we handled message}
    if Handled then
      {we did - return 0}
      Result := 0
    else
      {we didn't - pass message on to Windows}
      Result := DefWindowProc(fHWnd, Msg, WParam, LParam);
  end;
end;


procedure Register;
  {Component registration procedure}
begin
  RegisterComponents('DelphiDabbler', [TPJCBViewer]);
end;

end.
