{ ##
  @FILE                     UBDiffUtils.pas
  @COMMENTS                 Contains utility function needed in translation of C
                            code to Pascal to implement some C library calls.
  @PROJECT_NAME             BDiff
  @PROJECT_DESC             Outputs difference of binary files (Pascal
                            translation of original C code).
  @AUTHOR                   Original C program: Stefan Reuther <Streu@gmx.de>\
                            Pascal translation: Peter Johnson
                            <peter@delphidabbler.com>
  @WEBSITE                  Pascal versions: http://www.delphidabbler.com/
  @COPYRIGHT                Original C program: copyright  1999 by Stefan
                            Reuther.\
                            Pascal translation: copyright  2003 by Peter
                            Johnson.
  @LEGAL_NOTICE             Copying this program is allowed, as long as you
                            include original source code and Pascal translation
                            and document changes you made in a user-visible way
                            so people know they're using the new version, not
                            the original C code or Pascal translation.
  @HISTORY(
    @REVISION(
      @VERSION              1.0
      @DATE                 28/11/2003
      @COMMENTS             Original version.
    )
  )
}


unit UBDiffUtils;


interface


uses
  // Delphi
  Windows;


{ emulates C std lib malloc function }
function malloc(Size: Integer): Pointer;

{ emulates C std lib stdin value by returning Windows standard output handle }
function stdout: Integer;

{ emulates C std lib stdin value by returning Windows standard error handle }
function stderr: Integer;

{ emulates C std lib isprint function: does not support locales }
function isprint(Ch: Char): Boolean;

{ emulates C std lib perror function: displays given message followed by
  description of last Windows error }
procedure perror(const Msg: string);

{ cut down version of C std lib strtoul function that only supports base 10 }
function StrToULDec(const s: PChar; var endp: PChar): Longword;

{ helper function that returns the octal representation of the given byte as a
  string: note that the value has no leading '\', just the digits }
function ByteToOct(V: Byte): string;

{ writes binary data to a file: this is a C fwrite replacement }
procedure WriteBin(Handle: THandle; BufPtr: Pointer; Size: Integer);

{ writes a single character or string to a file: used in place of C's string or
  char output functions where string does not need formatting }
procedure WriteStr(Handle: THandle; const S: string);

{ writes a string built from format string and arguments to file: used in place
  of C's fprintf and printf functions (but doesn't handle \n etc }
procedure WriteStrFmt(Handle: THandle; const Fmt: string; Args: array of const);


implementation


uses
  // Delphi
  SysUtils;


{ helper function that returns the octal representation of the given byte as a
  string: note that the value has no leading '\', just the digits }
function ByteToOct(V: Byte): string;
var
  Idx: Integer;
  M: Byte;
begin
  Result := '';
  for Idx := 1 to 3 do
  begin
    M := V mod 8;
    V := V div 8;
    Result := Chr(M + Ord('0')) + Result;
  end;
end;

{ emulates C std lib malloc function }
function malloc(Size: Integer): Pointer;
begin
  GetMem(Result, Size);
end;

{ emulates C std lib stdin value by returning Windows standard output handle }
function stdout: Integer;
begin
  Result := GetStdHandle(STD_OUTPUT_HANDLE);
end;

{ emulates C std lib stdin value by returning Windows standard error handle }
function stderr: Integer;
begin
  Result := GetStdHandle(STD_ERROR_HANDLE);
end;

{ emulates C std lib isprint function: does not support locales }
function isprint(Ch: Char): Boolean;
begin
  Result := Ch in [#32..#126];
end;

{ emulates C std lib perror function: displays given message followed by
  description of last Windows error }
procedure perror(const Msg: string);
begin
  WriteStrFmt(stderr, '%s: %s'#13#10,
    [Msg, SysUtils.SysErrorMessage(Windows.GetLastError)]);
end;

{ cut down version of C std lib strtoul function that only supports base 10 }
function StrToULDec(const s: PChar; var endp: PChar): Longword;
begin
  endp := s;
  Result := 0;
  while endp^ in ['0'..'9'] do
  begin
    Result := 10 * Result + (Ord(endp^) - Ord('0'));
    inc(endp);
  end;
end;

{ writes binary data to a file: this is a C fwrite replacement }
procedure WriteBin(Handle: THandle; BufPtr: Pointer; Size: Integer);
var
  Dummy: DWORD;
begin
  Windows.WriteFile(Handle, BufPtr^, Size, Dummy, nil);
end;

{ writes a single character or string to a file: used in place of C's string or
  char output functions where string does not need formatting }
procedure WriteStr(Handle: THandle; const S: string);
begin
  WriteBin(Handle, PChar(S), Length(S));
end;

{ writes a string built from format string and arguments to file: used in place
  of C's fprintf and printf functions (but doesn't handle \n etc }
procedure WriteStrFmt(Handle: THandle; const Fmt: string; Args: array of const);
begin
  WriteStr(Handle, Format(Fmt, Args));
end;

end.
