{ ##
  @FILE                     UBPatch.pas
  @COMMENTS                 Main program logic.
  @PROJECT_NAME             BPatch
  @PROJECT_DESC             Applies binary patches created by BDiff (Pascal
                            translation of original C code).
  @AUTHOR                   Original C program: Stefan Reuther <Streu@gmx.de>\
                            Pascal translation: Peter Johnson
                            <peter@delphidabbler.com>
  @WEBSITE                  Pascal versions: http://www.delphidabbler.com/
  @COPYRIGHT                Original C program: copyright  1999 by Stefan
                            Reuther.\
                            Pascal translation: copyright  2003 by Peter
                            Johnson.
  @LEGAL_NOTICE             Copying this program is allowed, as long as you
                            include original source code and Pascal translation
                            and document changes you made in a user-visible way
                            so people know they're using the new version, not
                            the original C code or Pascal translation.
  @HISTORY(
    @REVISION(
      @VERSION              1.0
      @DATE                 28/11/2003
      @COMMENTS             Original Pascal translation of original bpatch.c.
    )
  )
}

unit UBPatch;


{
  PASCAL IMPLEMENTATION NOTES:
    + The Pascal translation is only suitable for use on Windows targets - it
      compiles to a Win32 console application and uses the Windows API.
    + Patch files read via shell redirection in the Pascal translation are read
      in pure binary mode, removing necessity to use --input command line switch
      to read patch file that is required in C implementation on MS-DOS/Windows
      systems.
    + When only one file is passed on command the program attempts to update the
      file in place but fails since it can't rename the temporary output file as
      the final since this files already exists (i.e. it is the input file). The
      C implementation also exhibits this behaviour.
}


interface


{ the program's main interface code: called from the project file }
procedure Main;


implementation


{
  ABOUT THIS IMPLEMENTATION

  This version of BPatch is a literal translation of Stefan Reuther's bpatch.c
  file from BPatch v0.2.

  The Pascal translation is on the left in the following code, while the
  equivalent original C code is displayed on the right.
}


// ++++++++++++++++++++++++++++++++++++ // ++++++++++++++++++++++++++++++++++ //
// +        PASCAL TRANSLATION        + // +            ORIGINAL C          + //
// ++++++++++++++++++++++++++++++++++++ // ++++++++++++++++++++++++++++++++++ //
                                        //
                                        //
{$IOCHECKS OFF}                         //
                                        //
{ define flag for 16 & 32 bit Windows } //
{$IFDEF WINDOWS}                        //
  {$DEFINE MSWINDOWS}                   //
{$ENDIF}                                //
{$IFDEF WIN32}                          //
  {$DEFINE MSWINDOWS}                   //
{$ENDIF}                                //
                                        //
                                        //
uses                                    //
  // Delphi                             //
  Windows, SysUtils,                    //#include <string.h>
                                        //#include <stdio.h>
                                        //#include <stdlib.h>
  // Project: special types & routines  //
  UBPatchUtils, UBPatchTypes;           //
                                        //
const                                   //
  VERSION_ = '0.2.1(p)'; {updated}      //#define VERSION "0.2"
  FORMAT_VERSION = '02';                //#define FORMAT_VERSION "02"
  BUFFER_SIZE = 4096;                   //#define BUFFER_SIZE 4096
                                        //
var                                     //
  progname: string;                     //char* progname;
  tempfile: string = '';                //char* tempfile = 0;
  tempfd: Integer = 0;                  //FILE* tempfd = 0;
                                        //
{ exit program with error message }     //
procedure error_exit(msg: PChar);       //void error_exit(char* msg)
begin                                   //
  fprintf(stderr, '%s: %s'#13#10,       //  fprintf(stderr, "%s: %s\n",
    [progname, msg]);                   //..  progname, msg);
  if tempfd > 0 then                    //  if(tempfd)
    fclose(tempfd);                     //    fclose(tempfd);
  if tempfile <> '' then                //  if(tempfile)
    SysUtils.DeleteFile(tempfile);      //    remove(tempfile);
  Halt(1);                              //  exit(1);
end;                                    //}
                                        //
{ compute simple checksum }             //
function checksum(data: PChar; len:     //long checksum(char* data, size_t len,
  size_t; l: Longint): Longint;         //..long l)
begin                                   //{
  while len <> 0 do                     //  while(len--) {
  begin                                 //
    Dec(len);                           //
    l := ((l shr 30) and 3)             //    l = ((l >> 30) & 3) | (l << 2);
      or (l shl 2);                     //
    l := l xor PShortInt(data)^;        //    l ^= *data++;
    Inc(data);                          //
  end;                                  //  }
  Result := l;                          //  return l;
end;                                    //}
                                        //
{ get 32-bit quantity from char array } //
function getlong(p: PChar): Longint;    //long getlong(char* p)
var                                     //
  q: PByte;                             //unsigned char* q = (unsigned char* )p;
  l: LongWord;                          //unsigned long l = *q++;
begin                                   //{
  q := PByte(p);                        //  unsigned char* q =
                                        //..  (unsigned char* )p;
  l := q^;  Inc(q);                     //  unsigned long l = *q++;
  l := l + 256 * q^;  Inc(q);           //  l += 256UL * *q++;
  l := l + 65536 * q^;  Inc(q);         //  l += 65536UL * *q++;
  l := l + 16777216 * q^;               //  l += 16777216UL * *q++;
  Result := l;                          //  return l;
end;                                    //}
                                        //
                                        //
{ copy data from one stream to another, //
  computing checksums }                 //
{ allows dest = 0 }                     //
procedure copy_data(src, dest: Integer; //void copy_data(FILE* src, FILE* dest,
  amount, check: Longint;               //..long amount, long check,
  src_is_patch: Integer);               //..int src_is_patch)
var                                     //
  chk: Longint;                         //long chk = 0;
  buffer: array[0..BUFFER_SIZE-1] of    //char buffer[BUFFER_SIZE];
    Char;                               //
  now: size_t;                          //size_t now = (amount > BUFFER_SIZE) ?
                                        //..BUFFER_SIZE : amount;
begin                                   //{
  chk := 0;                             //  long chk = 0;
                                        //  char buffer[BUFFER_SIZE];
  while amount <> 0 do                  //  while(amount) {
  begin                                 //
    if amount > BUFFER_SIZE then        //    size_t now = (amount >
      now := BUFFER_SIZE                //..    BUFFER_SIZE) ? BUFFER_SIZE :
    else                                //..    amount;
      now := amount;                    //
    if fread(@buffer, 1, now, src)      //    if(fread(buffer, 1, now, src)
      <> now then                       //..    != now) {
    begin                               //
      if feof(src) then                 //      if(feof(src))
      begin                             //
        if src_is_patch <> 0 then       //        if(src_is_patch)
          error_exit('Patch garbled - ' //          error_exit("Patch garbled -
            + 'unexpected end of data') //..          unexpected end of data");
        else                            //        else
          error_exit('Source file does '//          error_exit("Source file does
            + 'not match patch');       //..          not match patch");
      end                               //
      else                              //      else
      begin                             //
        if src_is_patch <> 0 then       //        if(src_is_patch)
          error_exit('Error reading '   //          error_exit("Error reading
            + 'patch file')             //..          patch file");
        else                            //        else
          error_exit('Error reading '   //          error_exit("Error reading
            + 'source file');           //..          source file");
      end;                              //
    end;                                //    }
    if dest <> 0 then                   //    if(dest)
      if fwrite(@buffer, 1, now,        //      if(fwrite(buffer, 1, now,
        dest) <> now then               //..      dest) != now)
        error_exit('Error writing '     //        error_exit("Error writing
          + 'temporary file');          //..        temporary file");
    chk := checksum(buffer, now, chk);  //    chk = checksum(buffer, now, chk);
    Dec(amount, now);                   //    amount -= now;
  end;                                  //  }
  if (src_is_patch = 0)                 //  if(!src_is_patch && chk != check)
    and (chk <> check) then             //
    error_exit('Source file does not '  //    error_exit("Source file does not
      + 'match patch');                 //      match patch");
end;                                    //}
                                        //
{ apply patch }                         //
procedure bpatch_(const src,            //void bpatch(const char* src,
  dest: PChar);                         //..const char* dest)
var                                     //
  sf: Integer; {source file}            //FILE* sf;
  df: Integer; {destination file}       //FILE* df;
  header: array[0..15] of Char;         //char header[16];
  p: PChar;                             //char* p;
  q: PChar;                             //char *q = strrchr(p, '\\');
  srclen, destlen: Longint;             //long srclen, destlen;
  size: Longint;                        //long size;
  ofs: Longint;                         //long ofs;
  c: Integer;                           //int c = fgetc(stdin);
const                                   //
  error_msg = 'Patch garbled - invalid '//static char error_msg[] = "Patch
    + 'section ''%''';                  //..garbled - invalid section `%'";
begin                                   //{
                                        //  FILE* sf;
                                        //  FILE* df;
                                        //  char header[16];
                                        //  char* p;
                                        //  long srclen, destlen;
                                        //  long size;
                                        //  long ofs;
  { read header }                       //
  if fread(@header, 1, 16,              //  if(fread(header, 1, 16,
    stdin) <> 16 then                   //..  stdin) != 16)
    error_exit('Patch not in BINARY '   //    error_exit("Patch not in BINARY
      + 'format');                      //..    format");
  if StrLComp(header, PChar('bdiff' +   //  if(strncmp(header, "bdiff"
    FORMAT_VERSION + #$1A), 8) <> 0 then//..  FORMAT_VERSION "\x1A", 8) != 0)
    error_exit('Patch not in BINARY '   //    error_exit("Patch not in BINARY
      + 'format');                      //..    format");
  srclen := getlong(@header[8]);        //  srclen = getlong(&header[8]);
  destlen := getlong(@header[12]);      //  destlen = getlong(&header[12]);
                                        //
  { open source file }                  //
  sf := FileOpen(src,                   //  sf = fopen(src, "rb");
      fmOpenRead + fmShareDenyNone);    //
  if sf <= 0 then                       //  if(!sf) {
  begin                                 //
    perror(src);                        //    perror(src);
    Halt(1);                            //    exit(1);
  end;                                  //  }
                                        //
  { create temporary file }             //
  if StrLen(dest) = 0 then              //  if(strlen(dest) == 0)
    error_exit('Empty destination file '//    error_exit("Empty destination file
      + 'name');                        //..    name");
                                        //  tempfile = malloc(strlen(dest) + 1);
  { we use Pascal long string:          //  if(!tempfile)
    no need to malloc space for it }    //    error_exit("Virtual memory
                                        //..    exhausted");
  { hack source file name to get a      //
    suitable temp file name }           //
  tempfile := dest;                     //  strcpy(tempfile, dest);
  p := StrRScan(PChar(tempfile), '/');  //  p = strrchr(tempfile, '/');
  if not Assigned(p) then               //  if(!p)
    p := PChar(tempfile)                //    p = tempfile;
  else                                  //  else
    Inc(p);                             //    p++;
{$IFDEF MSWINDOWS}                      //#ifdef __MSDOS__
                                        //  {
  q := StrRScan(p, '\');                //    char *q = strrchr(p, '\\');
  if Assigned(q) then                   //    if(q)
    p := q + 1;                         //      p = q + 1;
  q := StrRScan(p, ':');                //    q = strrchr(p, ':');
  if Assigned(q) then                   //    if(q)
    p := q + 1;                         //      p = q + 1;
                                        //  }
{$ENDIF}                                //#endif
  p^ := '$';                            //  *p = '$';
  df := FileCreate(tempfile);           //  df = fopen(tempfile, "wb");
  if df <= 0 then                       //  if(!df)
    error_exit('Can''t create '         //    error_exit("Can't create temporary
      + 'temporary file');              //..    file");
  tempfd := df;                         //  tempfd = df;
                                        //
  { apply patch }                       //
  while True do                         //  while(1) {
  begin                                 //
                                        //    static char error_msg[] = "Patch
                                        //..    garbled - invalid section `%'";
    c := fgetc(stdin);                  //    int c = fgetc(stdin);
    if c = EOF then                     //    if(c == EOF)
      Break;                            //      break;
    case c of                           //    switch(c) {
      Integer('@'):                     //      case '@':
      begin                             //
        { copy from source }            //
        if fread(@header, 1, 12,        //        if(fread(header, 1, 12,
          stdin) <> 12 then             //..        stdin) != 12)
          error_exit('Patch garbled - ' //          error_exit("Patch garbled -
            + 'unexpected end of data');//..          unexpected end of data");
        size := getlong(@header[4]);    //        size = getlong(&header[4]);
        ofs := getlong(@header[0]);     //        ofs = getlong(&header[0]);
        if (ofs < 0) or (size <= 0)     //        if(ofs < 0 || size <= 0 ||
          or (ofs > srclen)             //..        ofs > srclen ||
          or (size > srclen)            //..        size > srclen
          or (size+ofs > srclen) then   //          || size+ofs > srclen)
          error_exit('Patch garbled - ' //          error_exit("Patch garbled -
            + 'invalid change request');//..          invalid change request");
        if fseek(sf, ofs,               //        if(fseek(sf, ofs,
          SEEK_SET) <> 0 then           //..        SEEK_SET) != 0)
          error_exit('''fseek'' on '    //          error_exit("`fseek' on
            + 'source file failed');    //..          source file failed");
        copy_data(sf, df, size,         //        copy_data(sf, df, size,
          getlong(@header[8]), 0);      //..        getlong(&header[8]), 0);
        Dec(destlen, size);             //        destlen -= size;
                                        //        break;
      end;                              //
      Integer('+'):                     //      case '+':
      begin                             //
        { copy N bytes from patch }     //
        if fread(@header, 1, 4,         //        if(fread(header, 1, 4,
          stdin) <> 4 then              //..        stdin) != 4)
          error_exit('Patch garbled - ' //          error_exit("Patch garbled -
            + 'unexpected end of data');//..          unexpected end of data");
        size := getlong(@header[0]);    //        size = getlong(&header[0]);
        copy_data(stdin, df, size,      //        copy_data(stdin, df, size,
          0, 1);                        //..        0, 1);
        Dec(destlen, size);             //        destlen -= size;
                                        //        break;
      end;                              //
      else                              //      default:
      begin                             //
        fclose(sf);                     //        fclose(sf);
        fclose(df);                     //        fclose(df);
        StrRScan(error_msg, '%')^ :=    //        *strchr(error_msg, '%') = c;
          Char(c);                      //
        error_exit(error_msg);          //        error_exit(error_msg);
      end;                              //
    end;                                //    }
    if destlen < 0 then                 //    if(destlen < 0)
      error_exit('Patch garbled - patch'//      error_exit("Patch garbled -
        + ' file longer than announced' //..      patch file longer than
        + ' in header');                //..      announced in header");
  end;                                  //  }
  if destlen <> 0 then                  //  if(destlen)
    error_exit('Patch garbled - '       //    error_exit("Patch garbled -
      + 'destination file shorter than '//..    destination file shorter than
      + 'announced in header');         //..    announced in header");
                                        //
  fclose(sf);                           //  fclose(sf);
  fclose(df);                           //  fclose(df);
  tempfd := 0;                          //  tempfd = 0;
  if not RenameFile(tempfile, dest) then//  if(rename(tempfile, dest) != 0) {
  begin                                 //
    error_exit('Can''t rename temporary'//    error_exit("Can't rename temporary
      + ' file');                       //..    file");
  end;                                  //  }
  tempfile := '';                       //  tempfile = 0;
end;                                    //}
                                        //
{ help & exit }                         //
procedure help;                         //void help()
begin                                   //{
  fprintf(stdout,                       //  printf("%s: binary `patch' - apply
    '%s: binary ''patch'' - apply '     //..  binary patch\n"
    + 'binary patch'#13#10              //
    + #13#10                            //    "\n"
    + 'Usage: %s [options] old-file '   //    "Usage: %s [options] old-file
    + 'new-file [<patch-file]'#13#10    //..    new-file [<patch-file]\n"
    + #13#10                            //    "\n"
    + 'Valid options:'#13#10            //    "Valid options:\n"
    + ' -i FN --input=FN     Set input '//    " -i FN --input=FN     Set input
    + 'file name (instead of stdin)'    //..    file name (instead of stdin)\n"
    + #13#10                            //
    + ' -h    --help         Show this '//    " -h    --help         Show this
    + 'help screen'#13#10               //..    help screen\n"
    + ' -v    --version      Show '     //    " -v    --version      Show
    + 'version information'#13#10       //..    version information\n"
    + #13#10                            //    "\n"
    + '(c) copyright 1999 Stefan '      //      "(c) copyright 1999 Stefan
    + 'Reuther <Streu@gmx.de>'#13#10,   //..      Reuther <Streu@gmx.de>\n",
    [progname, progname]);              //      progname, progname);
  Halt(0);                              //  exit(0);
end;                                    //}
                                        //
{ version & exit }                      //
procedure version;                      //void version()
  function ExeDate: string;             //
  var                                   //
    H: Integer;                         //
    DOSDate: Integer;                   //
  begin                                 //
    H := FileOpen(ParamStr(0),          //
      fmOpenRead + fmShareDenyNone);    //
    try                                 //
      DOSDate := FileGetDate(H);        //
    finally                             //
      FileClose(H);                     //
    end;                                //
    Result := DateToStr(                //
      FileDateToDateTime(DOSDate));     //
  end;                                  //
  // NOTE: there is no Pascal equivalent//
  // of __DATE__ so we display update   //
  // date of program file rather than C //
  // version's compilation date         //
begin                                   //{
  fprintf(stdout, 'bpatch-%s %s '       //  printf("bpatch-" VERSION " ("
    + #13#10, [VERSION_, ExeDate]);     //..  __DATE__ ")\n");
  Halt(0);                              //  exit(0);
end;                                    //}
                                        //
{ control }                             //
procedure Main;                         //int main(int argc, char** argv)
var                                     //
  oldfn: string;                        //char* oldfn = 0;
  newfn: string;                        //char* newfn = 0;
  infn: string;                         //char* infn = 0;
  i: Integer;                           //int i;
  p: PChar;                             //char* p = argv[i] + 1;
                                        //char* p = argv[i] + 1;
  argv: PChar;                          //
  fp: Integer;                          //
begin                                   //{
  oldfn := '';                          //  char* oldfn = 0;
  newfn := '';                          //  char* newfn = 0;
  infn := '';                           //  char* infn = 0;
                                        //  int i;
  progname := ExtractFileName(          //  progname = argv[0];
    ParamStr(0));                       //
                                        //
  i := 1;                               //  for(i = 1; i < argc; i++)
  while i <= ParamCount do              //
  begin                                 //
    argv := PChar(ParamStr(i) + #0#0#0);//
    if argv[0] = '-' then               //    if(argv[i][0] == '-') {
    begin                               //
      if argv[1] = '-' then             //      if(argv[i][1] == '-') {
      begin                             //
        { long option }                 //
        p := argv + 2;                  //        char* p = argv[i]+2;
        if StrComp(p, 'help') = 0 then  //        if(strcmp(p, "help")==0)
          help                          //          help();
        else if StrComp(p,              //        else if(strcmp(p,
          'version') = 0 then           //..        "version")==0)
          version                       //          version();
        else if StrComp(p,              //        else if(strcmp(p, "input")==0)
          'input') = 0 then             //
        begin                           //
          Inc(i);                       //
          argv := PChar(ParamStr(i));   //
          if (argv^ = #0) then          //          if(!argv[++i]) {
          begin                         //
            fprintf(stderr,             //            fprintf(stderr, "%s:
              '%s: missing argument to '//..            missing argument to
              + '''--input'''#13#10,    //..            `--input'\n", progname);
              [progname]);              //
            Halt(1);                    //            return 1;
          end                           //          } else
          else                          //
            infn := argv;               //            infn=argv[i];
        end                             //
        else if StrLComp(p, 'input=',   //        else if(strncmp(p, "input=",
          6) = 0 then                   //..        6)==0)
          infn := p + 6                 //          infn=p + 6;
        else                            //        else {
        begin                           //
          fprintf(stderr,               //          fprintf(stderr, "%s: unknown
            '%s: unknown option '       //..          option `--%s'\n"
            + '''--%s'''#13#10          //
            + '%s: try ''%s --help'' '  //            "%s: try `%s --help' for
            + 'for more information'    //..            more information\n",
            + #13#10,                   //
            [progname, p, progname,     //            progname, p, progname,
            progname]);                 //..            progname);
          Halt(1);                      //          return 1;
        end;                            //        }
      end                               //      } else {
      else                              //
      begin                             //
        { short option }                //
        p := argv + 1;                  //        char* p = argv[i] + 1;
        while p^ <> #0 do               //        while(*p) {
        begin                           //
          case p^ of                    //          switch(*p) {
            'h':                        //            case 'h':
              help;                     //              help();
            'v':                        //            case 'v':
              version;                  //              version();
            'i':                        //            case 'i':
            begin                       //
              Inc(i);                   //
              argv :=                   //
                PChar(ParamStr(i));     //
              if argv^ = #0 then        //              if(!argv[++i]) {
              begin                     //
                fprintf(stderr,         //                fprintf(stderr, "%s:
                  '%s: missing argument'//..                missing argument to
                  + ' to ''-i'''#13#10, //..                `-i'\n", progname);
                  [progname]);          //
                Halt(1);                //                return 1;
              end                       //              } else
              else                      //
                infn := argv;           //                infn=argv[i];
                                        //              break;
            end                         //
            else                        //            default:
            begin                       //
              fprintf(stderr,           //              fprintf(stderr, "%s:
                '%s: unknown option '   //..              unknown option
                + '''-%s'''#13#10       //..              `-%c'\n"
                + '%s: try ''%s'        //                "%s: try `%s --help'
                + ' --help'' for more ' //..                for more
                + 'information'#13#10,  //..                information\n",
                [progname, p^, progname,//                progname, *p,
                progname]);             //..                progname, progname);
              Halt(1);                  //              return 1;
            end;                        //
          end;                          //          }
          Inc(p);                       //          p++;
        end;                            //        }
      end;                              //      }
    end                                 //    } else {
    else                                //
    begin                               //
      if oldfn = '' then                //      if(!oldfn)
        oldfn := ParamStr(i)            //        oldfn = argv[i];
      else if newfn = '' then           //      else if(!newfn)
        newfn := ParamStr(i)            //        newfn = argv[i];
      else                              //      else
        error_exit('Too many file names'//        error_exit("Too many file
          + ' on command line');        //..        names on command line");
    end;                                //    }
    Inc(i);                             //
  end;                                  //
                                        //
  if oldfn = '' then                    //  if(!oldfn) {
  begin                                 //
    fprintf(stderr, '%s: File name '    //    fprintf(stderr, "%s: File name
      + 'argument missing'#13#10        //..    argument missing\n"
      + '%s: try ''%s --help'' for '    //      "%s: try `%s --help' for more
      + 'more information'#13#10,       //..      information\n",
      [progname, progname, progname]);  //      progname, progname, progname);
    Halt(1);                            //    return 1;
  end;                                  //  }
                                        //
  if newfn = '' then                    //  if(!newfn)
    newfn := oldfn;                     //    newfn = oldfn;
                                        //
  if (infn <> '') and (infn <> '-') then//  if(infn && strcmp(infn, "-") != 0) {
  begin                                 //
    fp := FileOpen(infn, fmOpenRead or  //    if(!freopen(infn, "rb", stdin)) {
      fmShareDenyNone);                 //
    if fp <= 0 then                     //
    begin                               //
      perror(infn);                     //      perror(infn);
      Halt(1);                          //      return 1;
    end;                                //    }
    Windows.SetStdHandle(               //
      STD_INPUT_HANDLE, fp);            //
  end;                                  //  }
                                        //
  bpatch_(PChar(oldfn), PChar(newfn));  //  bpatch(oldfn, newfn);
                                        //
                                        //  return 0;
end;                                    //}

end.
