{ ##
  @FILE                     UBPatchUtils.pas
  @COMMENTS                 Contains utility function needed in translation of C
                            code to Pascal to implement some C library calls.
  @PROJECT_NAME             BPatch
  @PROJECT_DESC             Applies binary patches created by BDiff (Pascal
                            translation of original C code).
  @AUTHOR                   Original C program: Stefan Reuther <Streu@gmx.de>\
                            Pascal translation: Peter Johnson
                            <peter@delphidabbler.com>
  @WEBSITE                  Pascal versions: http://www.delphidabbler.com/
  @COPYRIGHT                Original C program: copyright  1999 by Stefan
                            Reuther.\
                            Pascal translation: copyright  2003 by Peter
                            Johnson.
  @LEGAL_NOTICE             Copying this program is allowed, as long as you
                            include original source code and Pascal translation
                            and document changes you made in a user-visible way
                            so people know they're using the new version, not
                            the original C code or Pascal translation.
  @HISTORY(
    @REVISION(
      @VERSION              1.0
      @DATE                 28/11/2003
      @COMMENTS             Original version.
    )
  )
}


unit UBPatchUtils;


interface


uses
  // Delphi
  Windows,
  // Project
  UBPatchTypes;


{ emulates C std lib fread function using Windows file handle }
function fread(ptr: Pointer; size: size_t; nobj: size_t;
  stream: Integer): size_t;

{ emulates C std lib fwrite function using Windows file handle }
function fwrite(ptr: Pointer; size: size_t; nobj: size_t;
  stream: Integer): size_t;

{ emulates C std lib fgetc function using Windows file handle }
function fgetc(stream: Integer): Integer;

{ emulates C std lib feof function using Windows file handle }
function feof(stream: Integer): Boolean;

{ emulates C std lib fseek function using Windows file handle }
function fseek(stream: Integer; offset: Longint; origin: Integer): Integer;

{ emulates C std lib fclose function using Windows file handle }
procedure fclose(stream: Integer);

{ emulates C std lib fprintf function using Windows file handle: \n etc not
  supported and parameters and string format specifiers are in Object Pascal
  format }
function fprintf(stream: Integer; const Fmt: string;
  Args: array of const): Integer;

{ emulates C std lib perror function: displays given message followed by
  description of last Windows error }
procedure perror(const Msg: string);

{ emulates C std lib stdin value by returning Windows standard input handle }
function stdin: Integer;

{ emulates C std lib stdin value by returning Windows standard output handle }
function stdout: Integer;

{ emulates C std lib stdin value by returning Windows standard error handle }
function stderr: Integer;


implementation


uses
  // Delphi
  SysUtils;


{ emulates C std lib fread function using Windows file handle }
function fread(ptr: Pointer; size: size_t; nobj: size_t;
  stream: Integer): size_t;
begin
  Result := size_t(SysUtils.FileRead(stream, ptr^, size * nobj)) div size;
end;

{ emulates C std lib fwrite function using Windows file handle }
function fwrite(ptr: Pointer; size: size_t; nobj: size_t;
  stream: Integer): size_t;
begin
  Result := size_t(SysUtils.FileWrite(stream, ptr^, size * nobj)) div size;
end;

{ emulates C std lib fgetc function using Windows file handle}
function fgetc(stream: Integer): Integer;
var
  Ch: Char;
begin
  if feof(stream) then
    Result := EOF
  else
  begin
    fread(@Ch,1, 1, stream);
    Result := Integer(Ch);
  end;
end;

{ emulates C std lib feof function using Windows file handle}
function feof(stream: Integer): Boolean;
var
  CurPos: Integer;
  Size: Integer;
begin
  CurPos := SysUtils.FileSeek(stream, 0, 1);
  Size := Windows.GetFileSize(stream, nil);
  Result := CurPos = Size;
end;

{ emulates C std lib fseek function using Windows file handle}
function fseek(stream: Integer; offset: Longint; origin: Integer): Integer;
begin
  if SysUtils.FileSeek(stream, offset, origin) = -1 then
    Result := -1
  else
    Result := 0;
end;

{ emulates C std lib fclose function using Windows file handle}
procedure fclose(stream: Integer);
begin
  SysUtils.FileClose(stream);
end;

{ emulates C std lib fprintf function using Windows file handle: \n etc not
  supported and parameters and string format specifiers are in Object Pascal
  format }
function fprintf(stream: Integer; const Fmt: string;
  Args: array of const): Integer;
var
  FormattedStr: string;
begin
  FormattedStr := Format(Fmt, Args);
  Result := Length(FormattedStr);
  fwrite(PChar(FormattedStr), Result, 1, stream);
end;

{ emulates C std lib perror function: displays given message followed by
  description of last Windows error }
procedure perror(const Msg: string);
begin
  fprintf(stderr, '%s: %s'#13#10,
    [Msg, SysUtils.SysErrorMessage(Windows.GetLastError)]);
end;

{ emulates C std lib stdin value by returning Windows standard input handle }
function stdin: Integer;
begin
  Result := Windows.GetStdHandle(STD_INPUT_HANDLE);
end;

{ emulates C std lib stdin value by returning Windows standard output handle }
function stdout: Integer;
begin
  Result := Windows.GetStdHandle(STD_OUTPUT_HANDLE);
end;

{ emulates C std lib stdin value by returning Windows standard error handle }
function stderr: Integer;
begin
  Result := Windows.GetStdHandle(STD_ERROR_HANDLE);
end;

end.
